from radcad.backends import Executor
import radcad.core as core

# Pathos re-writes the core code in Python rather than C, for ease of maintenance at cost of performance
from pathos.multiprocessing import ProcessPool


class ExecutorPathos(Executor):
    def execute_runs(self):
        with ProcessPool(self.engine.processes) as pool:
            result = pool.map(
                core._single_run_wrapper,
                [
                    (config, self.engine.raise_exceptions)
                    for config in self.engine._run_generator
                ],
            )
            pool.close()
            pool.join()
            pool.clear()
        return result



import copy
from functools import reduce, partial
import logging
import pickle
import traceback
from typing import Dict, List, Tuple


def _update_state(initial_state, params, substep, result, substate, signals, state_update_tuple):
    state, function = state_update_tuple
    if not state in initial_state:
        raise KeyError("Invalid state key in partial state update block")
    state_key, state_value = function(
        params, substep, result, substate, signals
    )
    if not state_key in initial_state:
        raise KeyError(
            "Invalid state key returned from state update function"
        )
    if state == state_key:
        return (state_key, state_value)
    else:
        raise KeyError(
            f"PSU state key {state} doesn't match function state key {state_key}"
        )


def _single_run(
    result: list,
    simulation: int,
    timesteps: int,
    run: int,
    subset: int,
    initial_state: dict,
    state_update_blocks: list,
    params: dict,
    deepcopy: bool,
    drop_substeps: bool,
):
    logging.info(f"Starting simulation {simulation} / run {run} / subset {subset}")

    initial_state["simulation"] = simulation
    initial_state["subset"] = subset
    initial_state["run"] = run + 1
    initial_state["substep"] = 0
    if not initial_state.get("timestep", False):
        initial_state["timestep"] = 0

    result.append([initial_state])

    for timestep in range(0, timesteps):
        previous_state: dict = (
            result[0][0].copy()
            if timestep == 0
            else result[-1][-1:][0].copy()
        )

        substeps: list = []
        substate: dict = previous_state.copy()

        for (substep, psu) in enumerate(state_update_blocks):
            substate: dict = (
                previous_state.copy() if substep == 0 else substeps[substep - 1].copy()
            )
            substate_copy = pickle.loads(pickle.dumps(substate, -1)) if deepcopy else substate.copy()
            substate["substep"] = substep + 1
            
            signals: dict = reduce_signals(
                params, substep, result, substate_copy, psu
            )

            updated_state = map(
                partial(_update_state, initial_state, params, substep, result, substate_copy, signals),
                psu["variables"].items()
            )
            substate.update(updated_state)
            substate["timestep"] = (previous_state["timestep"] + 1) if timestep == 0 else timestep + 1
            substeps.append(substate)

        substeps = [substate] if not substeps else substeps
        result.append(substeps if not drop_substeps else [substeps.pop()])
    return result


def single_run(
    simulation=0,
    timesteps=1,
    run=0,
    subset=0,
    initial_state={},
    state_update_blocks=[],
    params={},
    deepcopy: bool=True,
    drop_substeps: bool=False,
) -> Tuple[list, Exception, str]:
    result = []

    try:
        return (
            _single_run(
                result,
                simulation,
                timesteps,
                run,
                subset,
                initial_state,
                state_update_blocks,
                params,
                deepcopy,
                drop_substeps,
            ),
            None, # Error
            None, # Traceback
        )
    except Exception as error:
        trace = traceback.format_exc()
        print(trace)
        logging.warning(
            f"Simulation {simulation} / run {run} / subset {subset} failed! Returning partial results if Engine.raise_exceptions == False."
        )
        return (result, error, trace)


def _single_run_wrapper(args):
    run_args, raise_exceptions = args
    try:
        results, exception, traceback = single_run(*tuple(run_args))
        if raise_exceptions and exception:
            raise exception
        else:
            return results, {
                    'exception': exception,
                    'traceback': traceback,
                    'simulation': run_args.simulation,
                    'run': run_args.run,
                    'subset': run_args.subset,
                    'timesteps': run_args.timesteps,
                    'parameters': run_args.parameters,
                    'initial_state': run_args.initial_state,
                }
    except Exception as e:
        if raise_exceptions:
            raise e
        else:
            return [], e


def generate_parameter_sweep(params: Dict[str, List[any]]):
    param_sweep = []
    max_len = 0
    for value in params.values():
        if len(value) > max_len:
            max_len = len(value)

    for sweep_index in range(0, max_len):
        param_set = {}
        for (key, value) in params.items():
            param = (
                value[sweep_index]
                if sweep_index < len(value)
                else value[-1]
            )
            param_set[key] = param
        param_sweep.append(param_set)

    return param_sweep


def _add_signals(acc, a: Dict[str, any]):
    for (key, value) in a.items():
        if acc.get(key, None):
            acc[key] += value
        else:
            acc[key] = value
    return acc


def reduce_signals(params: dict, substep: int, result: list, substate: dict, psu: dict):
    policy_results: List[Dict[str, any]] = list(map(lambda function: function(params, substep, result, substate), psu["policies"].values()))

    result: dict = {}
    result_length = len(policy_results)
    if result_length == 0:
        return result
    elif result_length == 1:
        return pickle.loads(pickle.dumps(policy_results[0], -1))
    else:
        return reduce(_add_signals, policy_results, result)



try:
    import ray
except ImportError:
    _has_extension = False
else:
    _has_extension = True


if not _has_extension:
    raise Exception("Optional extension dependency Ray not installed")


from radcad.backends import Executor
import radcad.core as core


class ExecutorRay(Executor):
    @ray.remote
    def _proxy_single_run(args):
        return core._single_run_wrapper(args)
    
    def execute_runs(self):
        ray.init(num_cpus=self.engine.processes, ignore_reinit_error=True)
        futures = [
            ExecutorRay._proxy_single_run.remote((config, self.engine.raise_exceptions))
            for config in self.engine._run_generator
        ]
        return ray.get(futures)

class ExecutorRayRemote(Executor):
    @ray.remote
    def _proxy_single_run(args):
        return core._single_run_wrapper(args)

    def execute_runs(self):
        print(
            "Using Ray remote backend, please ensure you've initialized Ray using ray.init(address=***, ...)"
        )
        futures = [
            ExecutorRayRemote._proxy_single_run.remote((config, self.engine.raise_exceptions))
            for config in self.engine._run_generator
        ]
        return ray.get(futures)



from radcad.core import _single_run_wrapper, generate_parameter_sweep
from radcad.engine import Engine
from collections import namedtuple
import copy


RunArgs = namedtuple("RunArgs", [
    "simulation",
    "timesteps",
    "run",
    "subset",
    "initial_state",
    "state_update_blocks",
    "parameters",
    "deepcopy",
    "drop_substeps",
])
Context = namedtuple("Context", "simulation run subset timesteps initial_state parameters")


class Model:
    def __init__(self, initial_state={}, state_update_blocks=[], params={}):
        self.substeps = []
        self.state = {
            **copy.deepcopy(initial_state),
            'simulation': 0,
            'subset': 0,
            'run': 1,
            'substep': 0,
            'timestep': 0
        }
        self.initial_state = copy.deepcopy(initial_state)
        self.state_update_blocks = state_update_blocks
        self.params = copy.deepcopy(params)
        self.exceptions = []
        self._raise_exceptions = True
        self._deepcopy = True
        self._drop_substeps = False

    def __iter__(self):
        while True:
            param_sweep = generate_parameter_sweep(self.params)
            _params = param_sweep[0] if param_sweep else {}
            run_args = RunArgs(
                simulation = 0,
                timesteps = 1,
                run = 0,
                subset = 0,
                initial_state = copy.deepcopy(self.state),
                state_update_blocks = self.state_update_blocks,
                parameters = _params,
                deepcopy = self._deepcopy,
                drop_substeps = self._drop_substeps,
            )
            result, exception = _single_run_wrapper((run_args, self._raise_exceptions))
            if exception: self.exceptions.append(exception)
            self.substeps = result.pop()
            self.state = self.substeps[-1] if self.substeps else self.state
            yield self

    def __call__(self, **kwargs):
        self._raise_exceptions = kwargs.pop("raise_exceptions", True)
        self._deepcopy = kwargs.pop("deepcopy", True)
        self._drop_substeps = kwargs.pop("drop_substeps", False)
        return self


class Base:
    def __init__(self) -> None:
        self.results = []
        self.exceptions = []

    def run(self):
        raise NotImplementedError("Method run() not implemented for class that extends Base")


class Simulation(Base):
    def __init__(self, model: Model, timesteps=100, runs=1, **kwargs):
        self.model = model
        self.timesteps = timesteps
        self.runs = runs

        self.index = kwargs.pop("index", 0)
        self.experiment = Experiment(self)

        if kwargs:
            raise Exception(f"Invalid Simulation option in {kwargs}")

    def run(self):
        self.experiment.engine._run(experiment=self.experiment)
        self.results = self.experiment.results
        self.exceptions = self.experiment.exceptions
        return self.results


class Experiment(Base):
    """
    An Experiment.
    """

    def __init__(self, simulations=[], **kwargs):
        self.engine = kwargs.pop("engine", Engine())

        self.simulations = []

        # Add and validate simulations
        self.add_simulations(simulations)

        # Hooks
        self.before_experiment = kwargs.pop("before_experiment", None)
        self.after_experiment = kwargs.pop("after_experiment", None)
        self.before_simulation = kwargs.pop("before_simulation", None)
        self.after_simulation = kwargs.pop("after_simulation", None)
        self.before_run = kwargs.pop("before_run", None)
        self.after_run = kwargs.pop("after_run", None)
        self.before_subset = kwargs.pop("before_subset", None)
        self.after_subset = kwargs.pop("after_subset", None)

        if kwargs:
            raise Exception(f"Invalid Experiment option in {kwargs}")

    def run(self):
        return self.engine._run(experiment=self)

    def add_simulations(self, simulations):
        if not isinstance(simulations, list):
            simulations = [simulations]
        if any(not isinstance(sim, Simulation) for sim in simulations):
            raise Exception("Invalid simulation added")
        self.simulations.extend(simulations)

    def clear_simulations(self):
        cleared = True if self.simulations else False
        self.simulations = []
        return cleared

    def get_simulations(self):
        return self.simulations

    # Hooks
    def _before_experiment(self, experiment=None):
        if self.before_experiment:
            self.before_experiment(experiment=experiment)

    def _after_experiment(self, experiment=None):
        if self.after_experiment:
            self.after_experiment(experiment=experiment)

    def _before_simulation(self, simulation=None):
        if self.before_simulation:
            self.before_simulation(
                simulation=simulation
            )

    def _after_simulation(self, simulation=None):
        if self.after_simulation:
            self.after_simulation(
                simulation=simulation
            )

    def _before_run(self, context: Context=None):
        if self.before_run:
            self.before_run(context=context)

    def _after_run(self, context: Context=None):
        if self.after_run:
            self.after_run(context=context)

    def _before_subset(self, context: Context=None):
        if self.before_subset:
            self.before_subset(context=context)

    def _after_subset(self, context: Context=None):
        if self.after_subset:
            self.after_subset(context=context)



try:
    from cadCAD.configuration.utils import config_sim
except ImportError:
    _has_cadCAD = False
else:
    _has_cadCAD = True


if not _has_cadCAD:
    raise Exception("Optional compatibility dependency cadCAD not installed")


def config_sim(d):
    # radCAD handles configuration
    return d



def flatten(nested_list):
    def generator(nested_list):
        for sublist in nested_list:
            if isinstance(sublist, list):
                for item in sublist:
                    yield item
            else:
                yield sublist

    return list(generator(nested_list))


def extract_exceptions(results_with_exceptions):
    results, exceptions = zip(*results_with_exceptions)
    return (list(flatten(flatten(list(results)))), list(exceptions))



from radcad import Model, Simulation, Experiment
from radcad.engine import Engine, Backend

try:
    from cadCAD.engine import ExecutionMode, ExecutionContext
    import cadCAD.engine as _engine
except ImportError:
    _has_cadCAD = False
else:
    _has_cadCAD = True


if not _has_cadCAD:
    raise Exception("Optional compatibility dependency cadCAD not installed")


class Executor(_engine.Executor):
    def execute(self, engine=Engine()):
        simulations = []
        for config in self.configs:
            initial_state = config.initial_state
            state_update_blocks = config.partial_state_updates

            timesteps = max(list(config.sim_config["T"])) + 1
            runs = config.sim_config["N"]
            params = config.sim_config[
                "M"
            ]  # {key: [value] for key, value in config.sim_config['M'].items()}

            model = Model(
                initial_state=initial_state,
                state_update_blocks=state_update_blocks,
                params=params,
            )
            simulation = Simulation(model=model, timesteps=timesteps, runs=1)

            simulations.append(simulation)

        experiment = Experiment(simulations=simulations)
        experiment.engine = engine
        result = experiment.run()
        return result, None, None
